import time
import requests
import json
import uuid
import socket


RETRY_MODEL_LOADS_TIME = 10
RESTART_SLEEP_TIME = 15
CHECK_CALCULATE_STATUS_TIME = 2


class ServiceSolverError(Exception):
    pass


class ServiceTimeoutError(Exception):
    pass


def try_for_waitness(func):
    def wrapper(*arg, **kwargs):
        for _ in range(2):
            try:
                return func(*arg, **kwargs)
            except socket.error:
                time.sleep(20)
        raise ServiceTimeoutError
    return wrapper


def try_with_timeout_error(func):
    def wrapper(*arg, **kwargs):
        try:
            return func(*arg, **kwargs)
        except socket.error:
            raise ServiceTimeoutError

    return wrapper


def try_close_session(func):
    def wrapper(*arg, **kwargs):
        try:
            return func(*arg, **kwargs)
        except socket.error:
            pass

    return wrapper


class ServiceSolver:
    model_path = None
    port = None
    host = None
    token = None

    # Инициализация экземпляра сервисного солвера
    def __init__(self, model_path, host='localhost', port='60612', task_id=''):
        self.model_path = model_path
        self.host = host
        self.port = port
        self.token = str(task_id) + '_' + str(uuid.uuid4()) if task_id else str(uuid.uuid4())

    @try_with_timeout_error
    def load_model(self):
        # Проверка, готов ли сервис начать новую сессию
        # Проверка, готов ли сервис начать новую сессию
        self.waiting_for_service_readiness()
        headers = {
            'session_id': str(self.token),
            'async_model_loading': str(True)
        }
        with open(self.model_path, 'rb') as f:
            files = {'file': f}
            response = requests.post(f'http://{self.host}:{self.port}/api/models/upload', headers=headers, files=files)

        while response.text == 'Exception initialize. Try again.':
            with open(self.model_path, 'rb') as f:
                files = {'file': f}
                print(f'Error initializing model. Sleep {RETRY_MODEL_LOADS_TIME} sec and try again')
                time.sleep(RETRY_MODEL_LOADS_TIME)
                response = requests.post(f'http://{self.host}:{self.port}/api/models/upload', headers=headers,
                                         files=files)

        if response.text != 'Success initialize':
            print(f'Error initializing model: {response.text}')
            raise ServiceSolverError(f'Failed model initialization in service: {response.text}')

        print(f'Success initializing model')

    @try_for_waitness
    def waiting_for_service_readiness(self):
        # проверка готовности сервиса к работе
        response = requests.get(f'http://{self.host}:{self.port}/api/dwsim/get_service_state')

        # если сервису нужна перезагрузка - ждём
        while response.text == 'Restarting':
            print(f'Service waiting for reboot, sleep {RESTART_SLEEP_TIME} sec and check again')
            time.sleep(RESTART_SLEEP_TIME)
            response = requests.get(f'http://{self.host}:{self.port}/api/dwsim/get_service_state')

        # после рестарта сервис может перегружаться в течение максимум 20 секунд
        if response.text != 'Ready':
            time.sleep(20)
            response = requests.get(f'http://{self.host}:{self.port}/api/dwsim/get_service_state')

        # если после этого сервис не готов - бросаем исключение
        if response.text != 'Ready':
            print(f'Service unknown status: {response.text}')
            raise ServiceSolverError(f'Service unknown status: {response.text}')

        # сервис готов к работе
        print(f'Service is ready to work')

    @try_with_timeout_error
    def reload_flowsheet(self, model_path):
        data = {
            'session_id': str(self.token),
            'dwsim_model': model_path
        }
        requests.post(f'http://localhost:{self.port}/api/dwsim/reload_flowsheet', data=json.dumps(data))

    @try_with_timeout_error
    def set_property_value(self, object_name, property_name, property_value):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_name': property_name,
                    'prop_value': property_value
                }
            ]
        }
        requests.post(f'http://localhost:{self.port}/api/dwsim/set_property', data=json.dumps(data))

    @try_with_timeout_error
    def set_cell_value(self, worksheet_name, cell_name, cell_value):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'sheet': worksheet_name,
                    'prop_name': cell_name,
                    'prop_value': cell_value
                }
            ]
        }
        requests.post(f'http://localhost:{self.port}/api/dwsim/set_cell_value', data=json.dumps(data))

    @try_with_timeout_error
    def get_property_value(self, object_name, property_name):
        data = {
            'session_id': str(self.token),
            'output_data': [
                {
                    'name': object_name,
                    'prop_name': property_name
                }
            ]
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_property', data=json.dumps(data))
        return float(service_response.text)

    @try_with_timeout_error
    def get_cell_value(self, worksheet_name, cell_name, cast_to=str):
        data = {
            'session_id': str(self.token),
            'output_data': [
                {
                    'sheet': worksheet_name,
                    'prop_name': cell_name
                }
            ]
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_cell_value', data=json.dumps(data))
        return float(service_response.text) if cast_to == float else service_response.text

    @try_with_timeout_error
    def save_flowsheet(self, save_path, compressed):
        data = {
            'session_id': str(self.token),
            'save_path': save_path,
            'compressed': compressed
        }
        requests.get(f'http://localhost:{self.port}/api/dwsim/save_flowsheet',
                     data=json.dumps(data))

    @try_with_timeout_error
    def is_flowsheet_solved(self):
        data = {'session_id': str(self.token)}
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/is_flowsheet_solved',
                                        data=json.dumps(data))
        return True if service_response.text == 'True' else False

    @try_with_timeout_error
    def get_flowsheet_error_message(self):
        data = {'session_id': str(self.token)}
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_flowsheet_error_message',
                                        data=json.dumps(data))
        return str(service_response.text)

    @try_close_session
    def close_session(self):
        data = {'session_id': str(self.token)}
        requests.post(f'http://localhost:{self.port}/api/dwsim/session_id_close',
                      data=json.dumps(data))

    @try_with_timeout_error
    def calculate(self, attempts=1):
        # Send request
        print('Calculation starts')
        data = {
            'session_id': str(self.token),
            'attempts': attempts
        }
        service_response = requests.post(f'http://localhost:{self.port}/api/dwsim/calculate/',
                                         data=json.dumps(data))
        time.sleep(0.5)
        calc_id = service_response.text

        # Trying to get a result
        data2 = {
            'session_id': str(self.token)
        }
        service_response2 = requests.get(f'http://localhost:{self.port}/api/dwsim/calculate/{calc_id}',
                                         data=json.dumps(data2))
        while service_response2.text != 'True' and service_response2.text != 'False':
            time.sleep(0.5)
            service_response2 = requests.get(f'http://localhost:{self.port}/api/dwsim/calculate/{calc_id}',
                                             data=json.dumps(data2))
        if service_response2.text == 'True':
            print('Successful calculation!')
            return True

        print('Failed calculation!')
        return False

    @try_with_timeout_error
    def set_flash_spec(self, object_name, spec):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_value': spec
                }
            ]
        }
        requests.post(f'http://localhost:{self.port}/api/dwsim/set_flash_spec',
                      data=json.dumps(data))

    @try_with_timeout_error
    def calc_equilibrium(self, object_name, param_name, undefined_param):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_name': param_name,
                    'prop_value': undefined_param
                }
            ]
        }
        requests.post(f'http://localhost:{self.port}/api/dwsim/calc_equilibrium',
                      data=json.dumps(data))

    @try_with_timeout_error
    def get_object_q(self, object_name):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name
                }
            ]
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_object_q',
                                        data=json.dumps(data))
        return float(service_response.text)

    # COPY FROM MATERIAL

    # В проекте ACID get_object_properties запрашиваются у одного экземпляра ServiceSolver, и передаются в
    # set_object_properties другого экземпляра ServiceSolver, поэтому в ACID эти методы нельзя объединить в один
    # условный copy_from_material, который относится к одному конкретному экземпляру ServiceSolver
    # В OMEGA в отличие от ACID можно использовать copy_from_material в таком цельном виде,
    # так как вся работа ведется внутри одного и того же экземпляра ServiceSolver
    # def copy_from_material(self, object_name_from, object_name_to):
    #     properties = self.get_object_properties(object_name_from)
    #     self.set_object_properties(object_name_to, properties)

    # Для единообразия метод copy_from_material закомментирован, но если кому-то нужно получить такой-же эффект как при
    # вызове copy_from_material - используйте связку двух следующих методов как в закомментированном примере выше

    @try_with_timeout_error
    def set_object_properties(self, object_name, properties):
        data = {
            'session_id': str(self.token),
            'name': object_name
        }
        data = {**data, **json.loads(properties)}
        requests.post(f'http://localhost:{self.port}/api/dwsim/set_object_properties',
                      data=json.dumps(data))

    @try_with_timeout_error
    def get_object_properties(self, object_name):
        data = {
            'session_id': str(self.token),
            'name': object_name
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_object_properties',
                                        data=json.dumps(data))
        return service_response.text

    # ADD FROM ACID
    @try_with_timeout_error
    def is_graphic_object_active(self, object_name):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name
                }
            ]
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/is_graphic_object_active',
                                        data=json.dumps(data))
        return True if service_response.text == 'True' else False

    @try_with_timeout_error
    def set_ratios(self, object_name, ratios_index, value):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_name': ratios_index,
                    'prop_value': value
                }
            ]
        }
        requests.get(f'http://localhost:{self.port}/api/dwsim/set_ratios',
                     data=json.dumps(data))

    @try_with_timeout_error
    def set_stream_flow_spec(self, object_name, value):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_value': value
                }
            ]
        }
        requests.get(f'http://localhost:{self.port}/api/dwsim/set_stream_flowspec',
                     data=json.dumps(data))

    @try_with_timeout_error
    def get_input_variables(self, object_name, param_name):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_name': param_name
                }
            ]
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_input_variables',
                                        data=json.dumps(data))
        return float(service_response.text)

    @try_with_timeout_error
    def get_object_temperature(self, object_name, phase_index):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_value': phase_index
                }
            ]
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_object_temperature',
                                        data=json.dumps(data))
        return float(service_response.text)

    @try_with_timeout_error
    def get_object_pressure(self, object_name, phase_index):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_value': phase_index
                }
            ]
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_object_pressure',
                                        data=json.dumps(data))
        return float(service_response.text)

    @try_with_timeout_error
    def get_object_volumetric_flow(self, object_name, phase_index):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_value': phase_index
                }
            ]
        }
        service_response = requests.get(f'http://localhost:{self.port}/api/dwsim/get_object_volumetric_flow',
                                        data=json.dumps(data))
        return float(service_response.text)

    @try_with_timeout_error
    def set_object_volumetric_flow(self, object_name, phase_index, value):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name,
                    'prop_name': phase_index,
                    'prop_value': value
                }
            ]
        }
        requests.get(f'http://localhost:{self.port}/api/dwsim/set_object_volumetric_flow',
                     data=json.dumps(data))

    @try_with_timeout_error
    def calculate_object(self, object_name):
        data = {
            'session_id': str(self.token),
            'input_data': [
                {
                    'name': object_name
                }
            ]
        }
        requests.get(f'http://localhost:{self.port}/api/dwsim/calculate_object',
                     data=json.dumps(data))
